from .tests_tqdm import StringIO, closing, importorskip, mark, skip


@mark.slow
def test_rlock_creation():
    """Test that importing tqdm does not create multiprocessing objects."""
    mp = importorskip('multiprocessing')
    if not hasattr(mp, 'get_context'):
        skip("missing multiprocessing.get_context")

    # Use 'spawn' instead of 'fork' so that the process does not inherit any
    # globals that have been constructed by running other tests
    ctx = mp.get_context('spawn')
    with ctx.Pool(1) as pool:
        # The pool will propagate the error if the target method fails
        pool.apply(_rlock_creation_target)


def _rlock_creation_target():
    """Check that the RLock has not been constructed."""
    import multiprocessing as mp
    patch = importorskip('unittest.mock').patch

    # Patch the RLock class/method but use the original implementation
    with patch('multiprocessing.RLock', wraps=mp.RLock) as rlock_mock:
        # Importing the module should not create a lock
        from tqdm import tqdm
        assert rlock_mock.call_count == 0
        # Creating a progress bar should initialize the lock
        with closing(StringIO()) as our_file:
            with tqdm(file=our_file) as _:  # NOQA
                pass
        assert rlock_mock.call_count == 1
        # Creating a progress bar again should reuse the lock
        with closing(StringIO()) as our_file:
            with tqdm(file=our_file) as _:  # NOQA
                pass
        assert rlock_mock.call_count == 1



"""
Auto-generate README.rst from .meta/.readme.rst and docstrings.
"""
import re
import subprocess
import sys
from io import StringIO
from pathlib import Path
from textwrap import dedent

sys.path.insert(0, str(Path(__file__).parent.parent))
import tqdm  # noqa: E402
import tqdm.cli  # noqa: E402

HEAD_ARGS = """
Parameters
----------
"""
HEAD_RETS = """
Returns
-------
"""
HEAD_CLI = """
Extra CLI Options
-----------------
name  : type, optional
    TODO: find out why this is needed.
"""


def doc2rst(doc, arglist=True, raw=False, md2rst=True):
    """
    arglist  : bool, whether to create argument lists
    raw  : bool, ignores arglist and indents by 10 spaces
    md2rst  : bool, converts markdown to reStructuredText
    """
    if md2rst:
        doc = doc.replace('`', '``')
    if raw:
        doc = doc.replace('\n', '\n          ')
        doc = doc.replace('\n          \n', '\n\n')
    else:
        doc = dedent(doc)
        if arglist:
            doc = '\n'.join(i if not i or i[0] == ' ' else '* ' + i + '  '
                            for i in doc.split('\n'))
    return doc


src_dir = Path(__file__).parent.resolve()
README_rst = (src_dir / '.readme.rst').read_text("utf-8")
class_doc, init_doc = tqdm.tqdm.__doc__.split('\n\n', 1)
DOC_tqdm = doc2rst(class_doc + '\n', False, md2rst=False).replace('\n', '\n      ')
DOC_tqdm_init = doc2rst('\n' + init_doc)
DOC_tqdm_init_args = DOC_tqdm_init.partition(doc2rst(HEAD_ARGS))[-1].replace(
    '\n      ', '\n    ').replace('\n      ', '\n    ')
DOC_tqdm_init_args, _, DOC_tqdm_init_rets = DOC_tqdm_init_args.partition(doc2rst(HEAD_RETS))
DOC_cli = doc2rst(tqdm.cli.CLI_EXTRA_DOC).partition(doc2rst(HEAD_CLI))[-1]
DOC_tqdm_tqdm = {}
for i in dir(tqdm.tqdm):
    doc = getattr(tqdm.tqdm, i).__doc__
    if doc:
        DOC_tqdm_tqdm[i] = doc2rst(doc, raw=True, md2rst=False)

# special cases
DOC_tqdm_init_args = DOC_tqdm_init_args.replace(' *,', ' ``*``,')
DOC_tqdm_init_args = DOC_tqdm_init_args.partition('* gui  : bool, optional')[0]

README_rst = (
    README_rst.replace('{DOC_tqdm}', DOC_tqdm)
    .replace('{DOC_tqdm.tqdm.__init__.Parameters}', DOC_tqdm_init_args)
    .replace('{DOC_tqdm.cli.CLI_EXTRA_DOC}', DOC_cli)
    .replace('{DOC_tqdm.tqdm.__init__.Returns}', DOC_tqdm_init_rets))
for k, v in DOC_tqdm_tqdm.items():
    README_rst = README_rst.replace('{DOC_tqdm.tqdm.%s}' % k, v)

# python -m tqdm --help
sys.stdout, man = StringIO(), sys.stdout
try:
    tqdm.cli.main(argv=['--help'])
except SystemExit:
    sys.stdout, man = man, sys.stdout
man.seek(0)
# tail -n+5
for _ in range(4):
    man.readline()
# sed -r
man = man.read().replace('\\', '\\\\')
man = re.sub(r'^  (--.*)=<(.*)>  : (.*)$', r'\n\\\1=*\2*\n: \3.', man, flags=re.M)
man = re.sub(r'^  (--.*)  : (.*)$', r'\n\\\1\n: \2.', man, flags=re.M)
man = re.sub(r'  (-.*, )(--.*)  ', r'\n\1\\\2\n: ', man, flags=re.M)
# cat ".meta/.tqdm.1.md" -
man = (src_dir / '.tqdm.1.md').read_text() + man
# pandoc -s -t man
try:
    p = subprocess.Popen(['pandoc', '-s', '-t', 'man', '-V', 'pandoc-version='],  # nosec B603
                         stdin=subprocess.PIPE, stdout=subprocess.PIPE)
except FileNotFoundError:
    MAN_1 = None
else:
    MAN_1, _ = p.communicate(input=man.encode())

if __name__ == "__main__":
    (src_dir.parent / 'README.rst').write_text(README_rst, encoding='utf-8')
    if MAN_1:
        (src_dir.parent / 'tqdm' / 'tqdm.1').write_bytes(MAN_1)
    else:
        print("pandoc not found, skipping man page generation", file=sys.stderr)



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.
from functools import partial


class Comparison:
    """Running time of wrapped empty loops"""
    def __init__(self, length):
        try:
            from time import process_time
            self.time = process_time
        except ImportError:
            from time import clock
            self.time = clock
        self.iterable = range(1, 1 + int(length))

    def run(self, cls):
        pbar = cls(self.iterable)
        t0 = self.time()
        all(pbar)  # pylint: disable=pointless-statement
        t1 = self.time()
        return t1 - t0

    def run_by_name(self, method):
        return getattr(self, method.replace("-", "_"))()

    def no_progress(self):
        return self.run(lambda x: x)

    def tqdm_optimised(self):
        from tqdm import tqdm
        return self.run(partial(tqdm, miniters=6e5, smoothing=0))

    def tqdm(self):
        from tqdm import tqdm
        return self.run(tqdm)

    def alive_progress(self):
        from alive_progress import alive_bar

        class wrapper:
            def __init__(self, iterable):
                self.iterable = iterable

            def __iter__(self):
                iterable = self.iterable
                with alive_bar(len(iterable)) as bar:
                    for i in iterable:
                        yield i
                        bar()

        return self.run(wrapper)

    # def progressbar(self):
    #     from progressbar.progressbar import ProgressBar
    #     return self.run(ProgressBar())

    def progressbar2(self):
        from progressbar import progressbar
        return self.run(progressbar)

    def rich(self):
        from rich.progress import track
        return self.run(track)


# thorough test against no-progress
slow = Comparison(6e6)


def track_tqdm(method):
    return slow.run_by_name(method)


track_tqdm.params = ["tqdm", "tqdm-optimised", "no-progress"]
track_tqdm.param_names = ["method"]
track_tqdm.unit = "Seconds (lower is better)"

# quick test against alternatives
fast = Comparison(1e5)


def track_alternatives(library):
    return fast.run_by_name(library)


track_alternatives.params = ["rich", "progressbar2", "alive-progress", "tqdm"]
track_alternatives.param_names = ["library"]
track_alternatives.unit = "Seconds (lower is better)"
